async function compressAccurately(file, config) {
    if (!(file instanceof Blob)) {
        throw new Error('compressAccurately(): First arg must be a Blob object or a File object.');
    }
    if (typeof config !== 'object') {
        config = Object.assign({
            size: config,
        });
    }
    // 如果指定体积不是数字或者数字字符串，则不做处理
    config.size = Number(config.size);
    if (Number.isNaN(config.size)) {
        return file;
    }
    // 如果指定体积大于原文件体积，则不做处理；
    if (config.size * 1024 > file.size) {
        return file;
    }
    config.accuracy = Number(config.accuracy);
    if (!config.accuracy
        || config.accuracy < 0.8
        || config.accuracy > 0.99) {
        config.accuracy = 0.95; // 默认精度0.95
    }
    const resultSize = {
        max: config.size * (2 - config.accuracy) * 1024,
        accurate: config.size * 1024,
        min: config.size * config.accuracy * 1024,
    };
    const dataURL = await filetoDataURL(file);
    let originalMime = dataURL.split(',')[0].match(/:(.*?);/)[1]; // 原始图像图片类型
    let mime = 'image/jpeg';
    if (checkImageType(config.type)) {
        mime = config.type;
        originalMime = config.type;
    }
    const image = await dataURLtoImage(dataURL);
    const canvas = await imagetoCanvas(image, Object.assign({}, config));
    /**
     * 经过测试发现，blob.size与dataURL.length的比值约等于0.75
     * 这个比值可以同过dataURLtoFile这个方法来测试验证
     * 这里为了提高性能，直接通过这个比值来计算出blob.size
     */
    const proportion = 0.75;
    let imageQuality = 0.5;
    let compressDataURL;
    const tempDataURLs = [null, null];
    /**
     * HTMLCanvasElement.toBlob()以及HTMLCanvasElement.toDataURL()压缩参数
     * 的最小细粒度为0.01，而2的7次方为128，即只要循环7次，则会覆盖所有可能性
     */
    for (let x = 1; x <= 7; x++) {
        compressDataURL = await canvastoDataURL(canvas, imageQuality, mime);
        const CalculationSize = compressDataURL.length * proportion;
        // 如果到循环第七次还没有达到精确度的值，那说明该图片不能达到到此精确度要求
        // 这时候最后一次循环出来的dataURL可能不是最精确的，需要取其周边两个dataURL三者比较来选出最精确的；
        if (x === 7) {
            if (resultSize.max < CalculationSize || resultSize.min > CalculationSize) {
                compressDataURL = [compressDataURL, ...tempDataURLs]
                    .filter(i => i) // 去除null
                    .sort((a, b) => Math.abs(a.length * proportion - resultSize.accurate)
                        - Math.abs(b.length * proportion - resultSize.accurate))[0];
            }
            break;
        }
        if (resultSize.max < CalculationSize) {
            tempDataURLs[1] = compressDataURL;
            imageQuality -= 0.5 ** (x + 1);
        } else if (resultSize.min > CalculationSize) {
            tempDataURLs[0] = compressDataURL;
            imageQuality += 0.5 ** (x + 1);
        } else {
            break;
        }
    }
    const compressFile = await dataURLtoFile(compressDataURL, originalMime);
    // 如果压缩后体积大于原文件体积，则返回源文件；
    if (compressFile.size > file.size) {
        return file;
    }
    return compressFile;
}


function filetoDataURL(file) {
    return new Promise((resolve) => {
        const reader = new FileReader();
        reader.onloadend = e => resolve(e.target.result);
        reader.readAsDataURL(file);
    });
}

function checkImageType(type) {
    return ['image/png', 'image/jpeg', 'image/gif'].some(i => i === type);
}


function dataURLtoImage(dataURL) {
    return new Promise((resolve, reject) => {
        const img = new Image();
        img.onload = () => resolve(img);
        img.onerror = () => reject(new Error('dataURLtoImage(): dataURL is illegal'));
        img.src = dataURL;
    });
}


async function imagetoCanvas(image, config) {
    const myConfig = {...config};
    const cvs = document.createElement('canvas');
    const ctx = cvs.getContext('2d');
    let height;
    let width;
    for (const i in myConfig) {
        if (Object.prototype.hasOwnProperty.call(myConfig, i)) {
            myConfig[i] = Number(myConfig[i]);
        }
    }
    // 设置宽高
    if (!myConfig.scale) {
        width = myConfig.width || myConfig.height * image.width / image.height || image.width;
        height = myConfig.height || myConfig.width * image.height / image.width || image.height;
    } else {
        // 缩放比例0-10，不在此范围则保持原来图像大小
        const scale = myConfig.scale > 0 && myConfig.scale < 10 ? myConfig.scale : 1;
        width = image.width * scale;
        height = image.height * scale;
    }
    // 当顺时针或者逆时针旋转90时，需要交换canvas的宽高
    if ([5, 6, 7, 8].some(i => i === myConfig.orientation)) {
        cvs.height = width;
        cvs.width = height;
    } else {
        cvs.height = height;
        cvs.width = width;
    }
    // 设置方向
    switch (myConfig.orientation) {
        case 3:
            ctx.rotate(180 * Math.PI / 180);
            ctx.drawImage(image, -cvs.width, -cvs.height, cvs.width, cvs.height);
            break;
        case 6:
            ctx.rotate(90 * Math.PI / 180);
            ctx.drawImage(image, 0, -cvs.width, cvs.height, cvs.width);
            break;
        case 8:
            ctx.rotate(270 * Math.PI / 180);
            ctx.drawImage(image, -cvs.height, 0, cvs.height, cvs.width);
            break;
        case 2:
            ctx.translate(cvs.width, 0);
            ctx.scale(-1, 1);
            ctx.drawImage(image, 0, 0, cvs.width, cvs.height);
            break;
        case 4:
            ctx.translate(cvs.width, 0);
            ctx.scale(-1, 1);
            ctx.rotate(180 * Math.PI / 180);
            ctx.drawImage(image, -cvs.width, -cvs.height, cvs.width, cvs.height);
            break;
        case 5:
            ctx.translate(cvs.width, 0);
            ctx.scale(-1, 1);
            ctx.rotate(90 * Math.PI / 180);
            ctx.drawImage(image, 0, -cvs.width, cvs.height, cvs.width);
            break;
        case 7:
            ctx.translate(cvs.width, 0);
            ctx.scale(-1, 1);
            ctx.rotate(270 * Math.PI / 180);
            ctx.drawImage(image, -cvs.height, 0, cvs.height, cvs.width);
            break;
        default:
            ctx.drawImage(image, 0, 0, cvs.width, cvs.height);
    }
    return cvs;
}

async function canvastoDataURL(canvas, quality, type) {
    if (!quality) {
        quality = 0.92
    }
    if (!type) {
        type = 'image/jpeg'
    }

    if (!checkImageType(type)) {
        type = 'image/jpeg';
    }
    return canvas.toDataURL(type, quality);
}


async function dataURLtoFile(dataURL, type) {
    const arr = dataURL.split(',');
    let mime = arr[0].match(/:(.*?);/)[1];
    const bstr = atob(arr[1]);
    let n = bstr.length;
    const u8arr = new Uint8Array(n);
    while (n--) {
        u8arr[n] = bstr.charCodeAt(n);
    }
    if (checkImageType(type)) {
        mime = type;
    }
    return new Blob([u8arr], {
        type: mime,
    });
}

export {
    compressAccurately
}

export default {
    compressAccurately
}

